/*
 * Title:        sag_connectivity_plugins.hpp
 * Description:  C++ API for writing connectivity plugins
 * $Copyright (c) 2015-2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: sag_connectivity_plugins.hpp 366779 2020-02-05 10:42:04Z matj $
 */

/** @file
 *
 * Contains the headers needed to implement your own Connectivity Plug-ins. 
 *
 * See the com::softwareag::connectivity::AbstractTransport, com::softwareag::connectivity::AbstractCodec and 
 * com::softwareag::connectivity::Message classes.
 */

#ifndef _SAG_CONNECTIVITY_PLUGINS_HPP_
#define _SAG_CONNECTIVITY_PLUGINS_HPP_

#include <sag_connectivity_cpp.hpp>
#include <sag_plugin_logging.hpp>
#include <memory>
#include <mutex>

// must define __STDC_FORMAT_MACROS before first include of inttypes else printf macros won't be defined
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS 1
#endif
#include <inttypes.h>

namespace com {
namespace softwareag {
namespace _DATAT_INTERNAL_CPP_NAMESPACE { 

namespace
{
	/**
	 * Replace occurences of 'from' to 'to' in place in the string 'input'.
	 */
	void replace(std::string &input, const std::string &from, const std::string &to)
	{
		size_t pos = 0;
		while((pos = input.find(from, pos)) != std::string::npos)
		{
			input.replace(pos, from.length(), to);
			pos += to.length();
		}
	}

	/**
	 * Stringify a message. Intended only for logging, not performant.
	 */
	std::string to_string(const Message &m, bool truncate=true)
	{
		std::string payload = to_string(m.getPayload());
		if (truncate && payload.length() > 200) payload = payload.substr(0, 196) + " ...";
		// security sanitization to prevent log message faking
		replace(payload, "\n", "\\n");
		replace(payload, "\r", "\\r");
		return "Message<metadata="+to_string(m.getMetadataMap())+", payload="+payload+">";
	}
}

/**
 *  The enumeration indicating the direction of message flow - towards the transport or towards the host.
 *  
 *  Supports the 'less than' operator to allow its use in the key of a std::map, std::set, etc. 
 */
enum class Direction { 
	/**
	 * The direction of messages flowing towards the host (from the transport).
	 */
	TOWARDS_HOST = 1,
	/**
	 * The direction of messages flowing towards the transport (from the host).
	 */
	TOWARDS_TRANSPORT = 2

};

/** 
 * Interface to support miscellaneous requests from a particular plug-in to the host system.
 * @since 10.0
 */
class PluginHost {
	friend class Plugin;
public:
	typedef std::unique_ptr<PluginHost> ptr_t;

	/**
	 * Enable reliable messaging for the chain that this plug-in belongs to, in
	 * a particular direction i.e. TOWARDS_HOST means that the host is expected
	 * to acknowledge messages that this plug-in sends towards the host, and
	 * TOWARDS_TRANSPORT means that the transport is expected to acknowledge
	 * flushes for messages sent from the host.
	 *
	 * Must be called before this plug-in attempts to perform any other
	 * reliable messaging operations, and before the plug-in has been fully
	 * started.
	 *
	 * Should be called once for each direction if this chain is performing
	 * bi-directional reliable messaging.
	 */
	void enableReliability(Direction direction) { 
		if (SAG_ERROR_OK != sag_enable_reliability(chain, static_cast<int>(direction))) { 
			throw std::runtime_error("An error occurred while setting chain reliability"); 
		}
	}
	
	/** Check if host is shutting down. */
	bool isShuttingDown() { 
		bool isShuttingDown = false; 
		if (SAG_ERROR_OK != sag_is_host_shutting_down(chain, isShuttingDown)) {
			throw std::runtime_error("An error occurred while checking if host is shutting down");
		}
		return isShuttingDown; 
	}
private:
	/** Constructor - for internal use only
	 * @private
	 */
	PluginHost(void* chain = nullptr) :chain(chain) {}
	void* chain;
};

// forward decl for parameters friend
class Plugin;

/**
 * A base interface for parameters passed to the constructor of transport or codec plug-ins.
 *
 * Compatibility note: new methods may be added to this interface in future releases. 
 *
 * @see PluginConstructorParameters::TransportConstructorParameters
 * @see PluginConstructorParameters::CodecConstructorParameters
 * @since 9.12.0.1
 */
class PluginConstructorParameters
{
	// for constructor access to connectivityManager
	friend class Plugin;

public:
	class TransportConstructorParameters;

	class CodecConstructorParameters;

	/** Get the configuration for this plug-in. 
	 * 
	 * It is usually more convenient to access configuration using the non-const (mutable) config field instead of this method. 
	 */
	const map_t &getConfig() const { return config; }

	/**
	* Get the identifier used for the chain this plug-in is part of.
	*/
	const std::string &getChainId() const { return chainId; }

	/**
	* Get the name used in the configuration file for this plug-in.
	*/
	const std::string &getPluginName() const { return pluginName; }

protected:
	/** Constructor - for internal use only
	@private
	*/
	PluginConstructorParameters(const std::string &pluginName, const std::string &chainId, const map_t &config, void* connectivityManager, void* chain)
		: chainId(chainId), pluginName(pluginName), config(config.copy()), connectivityManager(connectivityManager), chain(chain)
	{}

private:
	std::string chainId;
	std::string pluginName;
	map_t config;
	/** For status reporting */
	void* connectivityManager;
	/** For creating PluginHost */
	void* chain;
};

/** A container for parameters passed to the constructor of a transport plug-in. */
class PluginConstructorParameters::TransportConstructorParameters : public PluginConstructorParameters
{
public:
	/** Constructor - for internal use only
	@private
	*/
	TransportConstructorParameters(const std::string &pluginName, const std::string &chainId, const map_t &config, void* connectivityManager, void* reserved)
	: PluginConstructorParameters(pluginName, chainId, config, connectivityManager, reserved)
	{}
};

/** A container for parameters passed to the constructor of a codec plug-in. */
class PluginConstructorParameters::CodecConstructorParameters : public PluginConstructorParameters
{
public:
	/** Constructor - for internal use only
	@private
	*/
	CodecConstructorParameters(const std::string &pluginName, const std::string &chainId, const map_t &config, void* connectivityManager, void* reserved)
	: PluginConstructorParameters(pluginName, chainId, config, connectivityManager, reserved)
	{}
	/** For initialization in epl plugin mode
	 @private
	*/
	CodecConstructorParameters(const std::string &pluginName)
	: PluginConstructorParameters(pluginName, "", map_t{}, nullptr, nullptr)
	{}
};


/**
 *  A class allowing a plug-in to report status values to the host.
 *  
 *  Ensure that no methods are called on this object after the plug-in has been shutdown.
 *  
 *  This object is NOT thread-safe, so please use additional locking if it is necessary to call its methods
 *  from multiple threads. However typically this will not be required, as usually all status items are created
 *  in a plug-in's constructor and each would be accessed subsequently from just one thread
 *  (e.g. hostwards status items on a thread handling hostwards
 *  messages and transportwards status items on a thread handling transportwards items).
 *  
 *  @since 10.0
 */
class StatusReporter
{
public:

	/** Returns a constant that should be used as the status value when a component is online, operational, connected,
	 *  and ready to handles messages.
	 */
	static const char* STATUS_ONLINE() { return "ONLINE"; }
	/** Returns a constant that should be used as the status value when a component is still starting, i.e. not yet online,
	 *  but not failed.
	 */
	static const char* STATUS_STARTING() { return "STARTING"; }
	/** Returns a constant that should be used as the status value when a component is not currently operational due to an
	 *  error condition.
	 */
	static const char* STATUS_FAILED() { return "FAILED"; }

	/**
	 * A class that can be used to efficiently update the value associated with a single status key.
	 * 
	 * This object is NOT thread-safe, so please use additional locking if it is necessary to update values from multiple threads.
	 * 
	 * Only one StatusItem instance can exist for a given key. Ensure that no methods
	 * are called on this object after the plug-in has been shutdown.
	 * 
	 */
	class StatusItem
	{
		friend class StatusReporter;
	public:

		~StatusItem()
		{
			sag_delete_user_status_item(connectivityManager, underlying);
			underlying.item = nullptr;
		}

		/** Set a string status value.
		 * 
		 * @param value the value to set e.g. STATUS_STARTING(), STATUS_ONLINE(), STATUS_FAILED().
		 * 
		 */
		void setStatus(const std::string &value)
		{
			std::unique_lock<std::mutex> ul(status_lock);
			setStatusLocked(value);
		}

		/** Set an integer status value.
		 *  
		 *  This is a convenience method that converts an integer value into a string automatically.
		 *  
		 *  This method takes a lock and may be called from multiple threads.
		 *  
		 *  @param value the value to set
		 */
		void setStatus(int64_t value)
		{
			std::unique_lock<std::mutex> ul(status_lock);
			intValue = value;
			setStatusLocked(convert_to_details::integerToString(value));
		}

		/** Return the value this status item was set to most recently by this 
		 *  class. 
		 *  
		 *  @return The last value this was set to.  
		 */
		std::string getStatus() { return lastValue; }

		/** Set an integer status value by incrementing the previous integer value that was set by this object.
		 * 
		 *  This method takes a lock and may be called from multiple threads.
		 * 
		 *  @param incrementValue the value to add on to the current value and set as the status. This can be a negative number.
		 *  Default value is +1.
		 */
		void increment(int64_t incrementValue = 1)
		{
			std::unique_lock<std::mutex> ul(status_lock);
			intValue += incrementValue;
			setStatusLocked(convert_to_details::integerToString(intValue));
		}

		/** Get the unique key specified when this status item was created
		*/
		const std::string &key() { return mkey; }

	private:
		StatusItem(const StatusItem& other) = delete; // non construction-copyable
		StatusItem& operator=(const StatusItem&) = delete; // non copyable

		StatusItem(void* connectivityManager, const std::string &key, const std::string &initialValue, const int64_t intValue)
			: intValue(intValue),
			mkey(key),
			lastValue(initialValue),
			connectivityManager(connectivityManager),
			underlying(sag_create_user_status_item(connectivityManager, key.c_str(), initialValue.c_str()))
		{
			if (!underlying.item)
			{
				std::ostringstream oss;
				oss << "Failed to create status item '" << key << "' (ensure the key is unique and that this plug-in has not been shutdown already)";
				throw std::runtime_error(oss.str());
			}
		}

		void setStatusLocked(const std::string &value)
		{
			if (value == lastValue) return; // no-op this case
			lastValue = value;

			sag_set_user_status_item(underlying, value.c_str());
		}

		int64_t intValue;
		const std::string mkey;
		std::string lastValue;
		void* connectivityManager;
		sag_status_item_t underlying;
		std::mutex status_lock;
	};

	/** Unique pointer to a StatusItem */
	typedef std::unique_ptr<StatusItem> item_ptr;

	/**
	 * Creates a StatusItem instance that can be used to report status for a given key.
	 * 
	 * It is an error to call this more than once for a given key.
	 * 
	 * The returned object must be deleted as part of plug-in destruction; the typical pattern for this is to
	 * store the unique_ptr as a member of the plug-in.
	 * 
	 * @param key a unique key that will identify this status item to the host.
	 * Typically this will include the chainId and plugin name, for example chainId+"."+pluginName+".messagesTowardsHost".
	 * @param initialValue the initial value for this item.
	 * @return The StatusItem instance that can be used to update the status value as needed; caller tasks ownership.
	 */
	item_ptr createStatusItem(const std::string &key, const std::string &initialValue)
	{
		return std::unique_ptr<StatusItem>(new StatusItem(connectivityManager, key, initialValue, 0));
	}

	/** Creates a StatusItem instance that can be used to report status for a given key, using an integral initial value.
	 *  
	 *  It is an error to call this more than once for a given key.
	 *  
	 *  The returned object must be deleted as part of plug-in destruction; the typical pattern for this is to
	 *  store the unique_ptr as a member of the plug-in.
	 *  
	 *  @param key a unique key that will identify this status item to the host.
	 *  Typically this will include the chainId and plugin name, for example chainId+"."+pluginName+".messagesTowardsHost".
	 *  @param initialValue the initial value for this item.
	 *  @return The StatusItem instance that can be used to update the status value as needed; caller tasks ownership.
	 */
	item_ptr createStatusItem(const std::string &key, int64_t initialValue)
	{
		std::ostringstream oss;
		oss << initialValue;
		return std::unique_ptr<StatusItem>(new StatusItem(connectivityManager, key, oss.str(), initialValue));
	}

	/** Set multiple related string status values at the same time (atomically).

	 *  The more efficient setStatus(const std::string &key, const std::string &value) should
	 *  be used instead unless there is a need to change multiple values atomically.
	 *  
	 *  Any status keys set using this method will be automatically removed when the plug-in instance is shutdown.
	 *  
	 *  @param statusmap A map containing data_t string (char*) keys and data_t string (char*) values.
	 *  Empty data_t() values can be used to indicate that an existing item should be cleared.
	 */
	void setStatus(const map_t &statusmap)
	{
		sag_set_user_status_map(connectivityManager, reinterpret_cast<const sag_underlying_map_t&>(statusmap));
		// keep track of what keys we've added so we can automatically remove them when we're destroyed
		for (auto it = statusmap.cbegin(); it != statusmap.cend(); it++)
			if (it->first.type_tag() == SAG_DATA_STRING) // ignore invalid ones
			{
			if (it->second.empty())
				mapKeysToCleanup.erase(it->first);
			else
				mapKeysToCleanup.insert(it->first.copy(), data_t());
			}
	}



	/** Destructor
	@private
	*/
	~StatusReporter()
	{
		if (!mapKeysToCleanup.empty())
			sag_set_user_status_map(connectivityManager, reinterpret_cast<const sag_underlying_map_t&>(mapKeysToCleanup));
	}

	/** @private
	For internal use only.
	*/
	explicit StatusReporter(void *connectivityManager) : connectivityManager(connectivityManager), mapKeysToCleanup()
	{
	}

private:
	void* connectivityManager;
	map_t mapKeysToCleanup;


	StatusReporter() = delete;

	// non-copyable, due to the cleanup code we don't want people to shoot themselves in the foot by having multiple copies
	StatusReporter(const StatusReporter& other) = delete; 
	StatusReporter& operator=(const StatusReporter&) = delete; 

};

/**
 * Base of the inheritance tree for Connectivity plugins.
 *
 * Normally plugins will implement either AbstractCodec or AbstractTransport, or one of their Simple variants.
 *
 * This base class provides several useful members to its derived classes. 
 */
class Plugin
{
public:
	/* Constructor.
	 * @since 9.12.0.1
	 */
	Plugin(const PluginConstructorParameters &params)
		: pluginName(params.getPluginName()), chainId(params.getChainId()), config(params.getConfig().copy()),
		host(new PluginHost(params.chain)), logger("connectivity." + pluginName + "." + chainId), 
		statusReporter(new StatusReporter(params.connectivityManager))
	{
	}

	/** This destructor must be virtual.
	 *
	 *  Destroy any objects used by the plug-in - no method on the plug-in will be called after the destructor has
	 *  begun.
	 */
	virtual ~Plugin() {}

	/** Called when an entire chain has been created and the plugin is allowed to start up (after all plugins are connected together).
	 * 
	 *  Some plug-ins will initiate a connection or create background threads in their start method. 
	 *  
	 *  You must not call methods on hostSide or transportSide before start() is called.
	 *  
	 *  The start method will never be called more than once on this instance.
	 *  
	 *  Default implementation is a no-op. 
	 */
	virtual void start() {}

	/** Called some time after start(),  when the host is ready to start receiving input (sends will be queued until this point).
	 *  
	 *  Plug-ins are not required to do anything at this point, though some transports may choose to avoid
	 *  pulling messages from an external system until the host is ready.
	 *  
	 *  Default implementation is a no-op.
	 */
	virtual void hostReady() {}

	/** Stop processing messages and terminate and join any background threads.
	 *
	 * The host will call this before shutdown.
	 *
	 * This method will never be called more than once on this instance. Implementations should ensure they work
	 * correctly even if the start() method was not called.
	 *
	 * This method may be called while messages are still being delivered towards the host or transport (it is
	 * even possible for messages to be delivered towards the transport after this method has returned - the plug-in
	 * should discard any messages sent to it after this point rather than attempt to process or deliver them). If
	 * the plug-in is blocked waiting to send a message, it should attempt to cancel the operation and unblock any
	 * threads sending messages transportwards when this method is called.  The plug-in must ensure that any calls
	 * to hostSide or transportSide are completed before returning from this method and not initiate any calls
	 * after this has completed.  This is typically achieved by interrupting and joining any threads that this
	 * plug-in has started (i.e. waiting for the threads to terminate).
	 *
	 * The plug-in must be left in a valid state - messages may be delivered to it after this call has completed;
	 * destroying objects and releasing can be done safely in the destructor.
	 *
	 * Default implementation is a no-op. You must override it if you need to wait for anything, and typically a
	 * transport will need to implement it to stop processing further messages.
	 */
	virtual void shutdown() {}
	/** The name used for this plug-in in the configuration file. */
	const std::string &getName() const { return pluginName; }
protected:
	/** The name used for this plug-in in the configuration file.
	This value is also available using the public getName() method. 
	@since 10.0 */
	const std::string pluginName;
	/** The identifier used for the chain this plug-in is part of. */
	const std::string chainId;
	/** The configuration of this plug-in. The plug-in owns and is free to mutate this data structure. */
	map_t config;
	/**
	 * Interface to support miscellaneous requests from this plug-in to the host system.
	 *
	 * Cannot be used with plug-ins that use the legacy constructor.
	 */
	const PluginHost::ptr_t host;
	/** Allows reporting status information from this plug-in, such as online or failed 
	 *  status and number of messages sent/received in each direction.
	 *  
	 *  Cannot be used with plug-ins that use the legacy constructor. 
	 *  
	 *  @since 10.0 
	 */
	StatusReporter &getStatusReporter() {
		if (statusReporter) return *statusReporter;
		throw std::runtime_error("Cannot call getStatusReporter when using the legacy constructor");
	}

public:
	/** Logging for writing to the host log file
	@since 9.12.0.1 
	*/
	Logger logger;
private:
	std::unique_ptr<StatusReporter> statusReporter;
};

/**
 * An interface to the next component (plugin or host) towards the host.
 * Contains the method to send messages towards the host.
 */
class HostSide
{
public:
	/** Ensure virtual destruction */
	virtual ~HostSide() {}
	/** Pointers to HostSides should always be this ptr_t type, which is a std::unique_ptr */
	typedef std::unique_ptr<HostSide> ptr_t;
	/** Abstract method that must be implemented to handle delivery of a batch of messages heading towards the host.
	 *
	 *  Messages should be processed and passed to the next connectivity plug-in in the chain, which is typically
	 *  provided by the Transport's or Codec's hostSide pointer. 
	 *
	 *  Implementors of this method are permitted to modify the contents of the messages (for example 
	 *  a codec may apply transformations to messages sent from a transport or host), though not to delete them. 
	 *  
	 *  The ability to handle messages in batches is provided to allow amortising the costs of delivering messages.
	 *  The batch is defined as all Messages from start up to but excluding end.  The callee may modify the
	 *  messages, typically by moving their contents elsewhere.
	 *  
	 *  Threading: Plug-ins are required to ensure that they do not call into the next plug-in in their chain from
	 *  more than one thread concurrently (although it does not have to be the same thread each time). Plug-ins may
	 *  therefore assume that no additional locking is required in this method to protect access to the plug-in's
	 *  state. The plug-in should be as responsive as possible to the Plugin.shutdown() method, which may be called
	 *  at any time on any thread, and should result in any calls into this unblocking and returning promptly. If
	 *  this method throws then the exception will be logged and the batch of messages dropped.
	 */
	virtual void sendBatchTowardsHost(Message *start, Message *end) = 0;

	/**
	 * Overload for sending a batch containing a single message.
	 *
	 * As sendBatchTowardsHost(Message*, Message*), but allows
	 * a single message to be sent.
	 * @since 10.2
	 */
	void sendBatchTowardsHost(Message &&message) {
		sendBatchTowardsHost(&message, &message+1);
	}

	/**
	 * Overload for sending messages using an iterator range.
	 *
	 * As sendBatchTowardsHost(Message*, Message*), but allows
	 * bidirectional iterators over Messages to be used.
	 * 
	 * Thus, can use <tt>std::vector<Message> msgs;</tt>
	 * <tt>sendBatchTowardsHost(msgs.begin(), msgs.end());</tt>
	 * @since 10.2
	 */
	template<typename IT>
	auto sendBatchTowardsHost(const IT &begin, const IT &end) -> typename std::enable_if<
		!std::is_const<ampl::remove_ref_t<decltype(*begin)>>::value &&
		ampl::is_same<Message, ampl::remove_const_t<ampl::remove_ref_t<decltype(*begin)>>>::value
		, void>::type // this ensures we can only pass in non-const iterator pairs to Message
	{
		if(begin == end) sendBatchTowardsHost((Message*) nullptr, (Message*) nullptr);
		else sendBatchTowardsHost(&(*begin), (&(*(end-1)))+1);
	}
};

/**
 * Wrap the next plugin in order as a HostSide object.
 * @private
 */
class RemoteHostSide: public HostSide
{
public:
	RemoteHostSide(sag_plugin_t other, sag_send_fn_t fn): other(other), fn(fn) {}
	virtual ~RemoteHostSide() {}
	virtual void sendBatchTowardsHost(Message *start, Message *end);
private:
	sag_plugin_t other;
	sag_send_fn_t fn;
};

/**
 * An interface to the next component (plugin or host) towards the transport.
 * Contains the method to send messages towards the transport.
 */
class TransportSide
{
public:
	/** Ensure virtual destruction */
	virtual ~TransportSide() {}
	/** Pointers to TransportSides should always be this ptr_t type, which is a std::unique_ptr */
	typedef std::unique_ptr<TransportSide> ptr_t;
	/** Abstract method that must be implemented to handle delivery of a batch of messages heading towards the transport.
	 *
	 * Messages should be processed and passed to the next connectivity plug-in in the chain, which is typically
	 * provided by the Codec's transportSide pointer.
	 *
	 *
	 * Implementors of this method are permitted to modify the contents of the messages (for example 
	 * a codec may apply transformations to messages sent from a transport or host), though not to delete them. 

	 * The ability to handle messages in batches is provided to allow amortising the costs of delivering messages.
	 * The batch is defined as all Messages from start up to but excluding end.  The callee may modify the
	 * messages, typically by moving their contents elsewhere.
	 *
	 * Threading: Plug-ins are required to ensure that they do not call into the next plug-in in their chain from
	 * more than one thread concurrently (although it does not have to be the same thread each time). Plug-ins may
	 * therefore assume that no additional locking is required in this method to protect access to the plug-in's
	 * state. The plug-in should be as responsive as possible to the Plugin.shutdown() method, which may be called
	 * at any time on any thread, and should result in any calls into this unblocking and returning promptly. If
	 * this method throws then the exception will be logged and the batch of messages dropped.
	 */
	virtual void sendBatchTowardsTransport(Message *start, Message *end) = 0;

	/**
	 * Overload for sending a batch containing a single message.
	 *
	 * As sendBatchTowardsTransport(Message*, Message*), but allows
	 * a single message to be sent.
	 * @since 10.2
	 */
	void sendBatchTowardsTransport(Message &&message) {
		sendBatchTowardsTransport(&message, &message+1);
	}

	/**
	 * Overload for sending messages using an iterator range.
	 *
	 * As sendBatchTowardsTransport(Message*, Message*), but allows
	 * bidirectional iterators over Messages to be used.
	 * Thus, can use <tt>std::vector<Message> msgs;</tt>
	 * <tt>sendBatchTowardsTransport(msgs.begin(), msgs.end());</tt>
	 * @since 10.2
	 */
	template<typename IT>
	auto sendBatchTowardsTransport(const IT &begin, const IT &end) -> typename std::enable_if<
		!std::is_const<ampl::remove_ref_t<decltype(*begin)>>::value &&
		ampl::is_same<Message, ampl::remove_const_t<ampl::remove_ref_t<decltype(*begin)>>>::value
		, void>::type // this ensures we can only pass in non-const iterator pairs to Message
	{
		if(begin == end) sendBatchTowardsTransport((Message*) nullptr, (Message*) nullptr);
		else sendBatchTowardsTransport(&(*begin), (&(*(end-1)))+1);
	}
};

/**
 * Wrap the next plugin in order as a TransportSide object.
 * @private
 */
class RemoteTransportSide: public TransportSide
{
public:
	RemoteTransportSide(sag_plugin_t other, sag_send_fn_t fn): other(other), fn(fn) {}
	virtual ~RemoteTransportSide() {}
	virtual void sendBatchTowardsTransport(Message *start, Message *end);
private:
	sag_plugin_t other;
	sag_send_fn_t fn;
};


/**
 *	Base class for codec plug-ins.
 *
 *	Implementors may derive directly from this class, providing an
 *	implementation of its abstract methods to provide any required
 *	transformation of Message objects as they are sent towards
 *	the transport or host side. For many codecs, inheriting from AbstractSimpleCodec
 *	may be more convenient than using this class directly, as it deals with batching of messages and
 *	ignoring messages with null payloads.
 *
 *	To export a codec class for use in the host you should use the SAG_DECLARE_CONNECTIVITY_CODEC_CLASS(Class) macro,
 *	and provide a public constructor with the same signature as
 *	AbstractCodec(const PluginConstructorParameters::CodecConstructorParameters &).
 *
 *	MapHelper can be used to extract the content of the message.
 *
 *	Derives from Plugin, which provides Plugin::pluginName, Plugin::chainId, Plugin::config and Plugin::logger members.
 *	Derives from HostSide, requiring derived classes to implement sendBatchTowardsHost(Message*, Message*)
 *	Derives from TransportSide, requiring derived classes to implement sendBatchTowardsTransport(Message*, Message*)
 *
 */
class AbstractCodec: public Plugin, public HostSide, public TransportSide
{
public:
	typedef PluginConstructorParameters::CodecConstructorParameters CodecConstructorParameters;

	/** Constructor. A subclass should provide its own constructor with the same signature as this one.
	 * 
	 *  A plug-in should throw an exception from the constructor if there is a problem with the configuration
	 *  or any other error that will prevent the plug-in from working correctly.
	 *  
	 *  @param params Provides access to the configuration for this plug-in, and other capabilities.
	 *  @since 9.12.0.1
	 */
	AbstractCodec(const PluginConstructorParameters::CodecConstructorParameters &params)
		: Plugin(params)
	{}

	// These methods do not need to show up in doxygen
	/* Called between construction and start() to set the hostSide field  */
	virtual void setNextTowardsHost(HostSide::ptr_t &&host)
	{
		hostSide = std::move(host);
	}
	/* Called between construction and start() to set the transportSide field */
	virtual void setNextTowardsTransport(TransportSide::ptr_t &&transport)
	{
		transportSide = std::move(transport);
	}
protected:
	/** The next plug-in in the chain towards host.
	 * 
	 *  This field will be given a valid value before the start() method is called.
	 */
	HostSide::ptr_t hostSide;

	/** The next plug-in in the chain towards transport.
	 *
	 *  This field will be given a valid value before the start() method is called.
	 */
	TransportSide::ptr_t transportSide;
};

/**
 *	Base class for transport plug-ins.
 *
 *	Implementors may derive directly from this class if you need to handle messages in batches, 
 *	typically by providing an implementation of sendBatchTowardsTransport(Message*,Message*), 
 *	and in the other direction sending messages towards the host from a background thread created by the start() method. 
 *	Messages are delivered to or from this class in batches, which transports may be able to make use of to deliver many 
 *	messages in a single operation, amortizing the cost of delivering messages. 
 *	The division into batches is of no significance beyond them all being available for delivery within a short space of time - 
 *	a transport should function the same if given the messages individually or in batches.
 *
 *	For many transport plug-ins, inheriting from AbstractSimpleTransport
 *	may be more convenient than using this class directly, as it deals with batching of messages automatically.
 *
 *	To export a transport class for use in the host you should use the SAG_DECLARE_CONNECTIVITY_TRANSPORT_CLASS(Class) macro,
 *	and provide a public constructor with the same signature as
 *	AbstractTransport(const PluginConstructorParameters::TransportConstructorParameters &params)
 *
 *	Derives from Plugin, which provides Plugin::pluginName, Plugin::chainId, Plugin::config and Plugin::logger members.
 *	Derives from TransportSide, requiring derived classes to implement sendBatchTowardsTransport(Message*, Message*)
 *
 *	Note: Some transports may need to respond to every message received from the host with a corresponding reply back towards the host. This has
 *	to be done carefully; if the implementation of TransportSide.sendBatchTowardsTransport() contains a simple
 *	synchronous call back to HostSide.sendBatchTowardsHost(), this has the potential to cause a deadlock in the host
 *	application, depending on the disposition of internal queues. Transports needing to follow this pattern should generate their responses
 *	on a distinct thread, not the thread that processes the message(s) from the host.
 *
 */
class AbstractTransport: public Plugin, public TransportSide
{
public:
	typedef PluginConstructorParameters::TransportConstructorParameters TransportConstructorParameters;

	/** Constructor. A subclass should provide its own constructor with the same signature as this one.
	 *  
	 *  A plug-in should throw an exception from the constructor if there is a problem with the configuration
	 *  or any other error that will prevent the plug-in from working correctly.
	 *  
	 *  @param params Provides access to the configuration for this plug-in, and other capabilities.
	 *  @since 9.12.0.1
	 */
	AbstractTransport(const PluginConstructorParameters::TransportConstructorParameters &params)
		: Plugin(params)
	{}

	// This method does not need to show up in doxygen
	/* Called between construction and start() to set the hostSide field */
	virtual void setNextTowardsHost(HostSide::ptr_t &&host)
	{
		hostSide = std::move(host);
	}

protected:
	/** The next plug-in in the chain towards host.
	 *
	 * This field will be given a valid value before the start() method is called.
	 */
	HostSide::ptr_t hostSide;
};

/** Base class that simplifies implementation of transport plug-ins that deal only with individual messages not batches, and 
 *	ignore messages with null payloads.
 *
 *	To export a transport class for use in the host you should use the SAG_DECLARE_CONNECTIVITY_TRANSPORT_CLASS(Class) macro,
 *	and provide a public constructor with the same signature as
 *	AbstractSimpleTransport(const PluginConstructorParameters::TransportConstructorParameters &).
 *
 *	Plug-in authors will typically extend this class by providing at least valid constructor, an implementation of 
 *	deliverMessageTowardsTransport(Message&), and sending messages towards the hostSide from a background thread 
 *	created in start(). 
 */
class AbstractSimpleTransport: public AbstractTransport
{
public:

	/** Constructor. A subclass should provide its own constructor with the same signature as this one.
	 * 
	 *  A plug-in should throw an exception from the constructor if there is a problem with the configuration
	 *  or any other error that will prevent the plug-in from working correctly.
	 *  
	 *  @param params Provides access to the configuration for this plug-in, and other capabilities.
	 *  @since 9.12.0.1
	 */
	AbstractSimpleTransport(const PluginConstructorParameters::TransportConstructorParameters &params)
		: AbstractTransport(params)
	{}


	/** Implements batch sending, calling deliverMessageTowardsTransport(Message&) for each message individually.
	 *
	 *	Any exceptions from deliverMessageTowardsTransport are caught and handleException(Message&) is called and then 
	 *	processing continues for the next message. 
	 */
	virtual void sendBatchTowardsTransport(Message *start, Message *end)
	{
		for (Message *it = start; it != end; ++it) {
			try {
				if (it->getPayload().empty()) {
					deliverNullPayloadTowardsTransport(*it);
				} else {
					deliverMessageTowardsTransport(*it);
				}
			} catch (...) {
				handleException(*it);
			}
		}
	}
	/** Abstract method that must be implemented to handle delivery of an individual message. */
	virtual void deliverMessageTowardsTransport(Message &msg) = 0;
	/** Deliver a message with a null payload. Default is to ignore it */
	virtual void deliverNullPayloadTowardsTransport(Message &msg) 
	{
		// do nothing
	}

	/** Handle an exception thrown while delivering a message towards the transport.
	 *
	 *  Default implementation is to log a warning and discard a message.
	 *  
	 *  You can override it yourself to have other behaviour.
	 *  The Message which caused the error is passed in.
	 *  
	 *  This is called from within a catch block, so if you need
	 *  to know the exception thrown then you can rethrow it yourself:
	 *  
	 *  \code
	 *  try {
	 *	 	throw;
	 *	 } catch (const type &ex) {
	 *	 	...
	 *	 }
	 *  \endcode
	 */
	virtual void handleException(Message &m)
	{
		try {
			throw;
		} catch (const std::exception &e) {
			logger.warn("Error while delivering message: %s; %s will be dropped.", e.what(), to_string(m).c_str());
		} catch (...) {
			logger.warn("Unknown error delivering message: %s", to_string(m).c_str());
		}
	}
};

/**
 *	Base class that simplifies implementation of codec plug-ins that deal only with individual messages not batches, and 
 *	ignore messages with null payloads.
 *
 *	To export a codec class for use in the host you should use the SAG_DECLARE_CONNECTIVITY_CODEC_CLASS(Class) macro,
 *	and provide a public constructor with the same signature as
 *	AbstractSimpleCodec(const PluginConstructorParameters::CodecConstructorParameters &).
 *
 *	Handles batches of messages and ignoring null payload messages (by default). Implementors will typically subclass
 *	this class, providing an implementation of the transformMessageTowardsHost(Message&) and
 *	transformMessageTowardsTransport(Message&) methods.
 *
 *	If you don't need to handle messages a batch at a time, then you should derive from this class.
 *	If you do then you should directly derive from AbstractCodec.
 */
class AbstractSimpleCodec: public AbstractCodec
{
public:

	/** Constructor. A subclass should provide its own constructor with the same signature as this one.
	 *  
	 *  A plug-in should throw an exception from the constructor if there is a problem with the configuration
	 *  or any other error that will prevent the plug-in from working correctly.
	 *  
	 *  @param params Provides access to the configuration for this plug-in, and other capabilities.
	 *  @since 9.12.0.1
	 */
	AbstractSimpleCodec(const PluginConstructorParameters::CodecConstructorParameters &params)
		: AbstractCodec(params)
	{}

	/** Implements batch sending, calling transformMessageTowardsHost(Message &) for each message individually.
	 *	Exceptions from transformMessageTowardsHost are caught and handleException(Message &, bool) is called
	 *	(the default implementation is to log a warning and discard the event)
	 *	If handleException or transformMessageTowardsHost returns false then that event is removed,
	 *	but processing continues for the next message. 
	 */
	virtual void sendBatchTowardsHost(Message *start, Message *end)
	{
		Message *curr = start;
		for (Message *it = start; it != end; ++it) {
			bool rv;
			try {
				if (it->getPayload().empty()) {
					rv = transformNullPayloadTowardsHost(*it);
				} else {
					rv = transformMessageTowardsHost(*it);
				}
			} catch (...) {
				rv = handleException(*it, false);
			}
			// if we keep it (and it didn't throw) swap it with the accumulator
			if (rv) {
				if (it != curr) it->swap(std::move(*curr)); // don't self-swap
				++curr;
			}
		}
		if (hostSide.get()) hostSide->sendBatchTowardsHost(start, curr);
	}
	/** Implements batch sending, calling transformMessageTowardsTransport(Message &) for each message individually.
	 *	Exceptions from transformMessageTowardsTransport are caught and handleException(Message &, bool) is called
	 *	(the default implementation is to log a warning and discard the event). 
	 *	If handleException or transformMessageTowardsTransport returns false then that event is removed,
	 *	but processing continues for the next message. 
	 */
	virtual void sendBatchTowardsTransport(Message *start, Message *end)
	{
		Message *curr = start;
		for (Message *it = start; it != end; ++it) {
			bool rv;
			try {
				// process the message
				if (it->getPayload().empty()) {
					rv = transformNullPayloadTowardsTransport(*it);
				} else {
					rv = transformMessageTowardsTransport(*it);
				}
			} catch (...) {
				rv = handleException(*it, true);
			}
			// if we keep it (and it didn't throw) swap it with the accumulator
			if (rv) {
				if (it != curr) it->swap(std::move(*curr)); // don't self-swap
				++curr;
			}
		}
		if (transportSide.get()) transportSide->sendBatchTowardsTransport(start, curr);
	}

	/** Abstract method that must be implemented to handle transformation of an individual message.
	 *	in a host-wards direction
	 *	@return true to keep the message and false (or throw) to delete it 
	 */
	virtual bool transformMessageTowardsHost(Message &msg) = 0;
	/** Abstract method that must be implemented to handle transformation of an individual message 
	 *	in a transport-wards direction. 
	 *	@return true to keep the message and false (or throw) to delete it 
	 */
	virtual bool transformMessageTowardsTransport(Message &msg) = 0;
	/** Transform a message with a null payload in a host-wards direction.
	 *	default is to do nothing
	 *	@return true to keep the message and false (or throw) to delete it 
	 */
	virtual bool transformNullPayloadTowardsHost(Message &msg)
	{
		// do nothing
		return true;
	}
	/** Transform a message with a null payload in a transport-wards direction.
	 *	default is to do nothing
	 *	@return true to keep the message and false (or throw) to delete it 
	 */
	virtual bool transformNullPayloadTowardsTransport(Message &msg)
	{
		// do nothing
		return true;
	}
	/** Handle an exception thrown while delivering a message.
	 * Default is to log a warning and discard a message.
	 *
	 * This is called from within a catch block, so if you need
	 * to know the exception thrown then you can rethrow it yourself:
	 *
	 * \code
	 * try {
	 *		throw;
	 * } catch (const type &ex) { ...
	 * }
	 * \endcode
	 *
	 * @param m The Message that caused the error is passed in by reference.
	 * @param towardsTransport set to true if it was
	 * transformMessageTowardsTransport that threw or false if it was
	 * transformMessageTowardsHost that threw
	 * @returns If you wish to fix the message and then continue passing it on
	 * then return true, if it should be discarded
	 * then should return false
	 */
	virtual bool handleException(Message &m, bool towardsTransport)
	{
		try {
			throw;
		} catch (const std::exception &e) {
			logger.warn("Error while transforming message: %s; %s will be dropped.", e.what(), to_string(m).c_str());
		} catch (...) {
			logger.warn("Unknown error transforming message: %s", to_string(m).c_str());
		}
		return false;
	}
};

}

namespace connectivity { using namespace _DATAT_INTERNAL_CPP_NAMESPACE; }

}} // com.softwareag.connectivity

// internal implementation included from these files
#include <sag_internal/exception.hpp>
#include <sag_internal/remote_plugins.hpp>
#include <sag_internal/plugin_macros.hpp>

/** Registers a connectivity transport plug-in class, by exporting the necessary C free functions from this library.
 *
 *	For each transport class you have (derived from com::softwareag::connectivity::AbstractTransport or com::softwareag::connectivity::AbstractSimpleTransport) you should call this macro once,
 *	and ensure that your class has a constructor with the same signature as 
 *	com::softwareag::connectivity::AbstractTransport(com::softwareag::connectivity::PluginConstructorParameters::TransportConstructorParameters &).
 *
 *	If your class is in a namespace then you must call this macro in the same namespace, passing in just the base name of the class.
 *	@since 9.12.0.1
 */
#define SAG_DECLARE_CONNECTIVITY_TRANSPORT_CLASS(Class) _SAG_DECLARE_CONNECTIVITY_TRANSPORT_CLASS(Class)

/** Registers a connectivity codec plug-in class, by exporting the necessary C free functions from this library.
 *
 *	For each codec class you have (derived from com::softwareag::connectivity::AbstractCodec or com::softwareag::connectivity::AbstractSimpleCodec) you should call this macro once, 
 *	and ensure that your class has a constructor with the same signature as 
 *	com::softwareag::connectivity::AbstractCodec(com::softwareag::connectivity::PluginConstructorParameters::CodecConstructorParameters &).
 *
 *	If your class is in a namespace then you must call this macro in the same namespace, passing in just the base name of the class.
 *	@since 9.12.0.1
 */
#define SAG_DECLARE_CONNECTIVITY_CODEC_CLASS(Class) _SAG_DECLARE_CONNECTIVITY_CODEC_CLASS(Class)

#endif // _SAG_CONNECTIVITY_PLUGINS_HPP_
